# Copyright (c) 2010 Infrae. All rights reserved.
# See also LICENSE.txt
# $Id$

from infrae.rest.components import REST, RESTWithTemplate, lookupREST
from infrae.rest.interfaces import IRESTMethodPublishedEvent

__all__ = [
    'REST', 'lookupREST', 'RESTWithTemplate', 'IRESTMethodPublishedEvent']




"""

  Let's first grok our test first:

   >>> grok('infrae.rest.tests.grok.absoluteurl')

  Now, if we have a folder we can use our REST component:

    >>> root = getRootFolder()
    >>> root.manage_addProduct['OFS'].manage_addFolder('folder', 'Folder')

  If we access it, it should return its URL:

    >>> print http('GET /folder/++rest++url HTTP/1.0')
    HTTP/1.0 200 OK
    Content-Length: 35
    Content-Type: text/html;charset=utf-8
    <BLANKLINE>
    http://localhost/folder/++rest++url

  That also works for nest components:

    >>> print http('GET /folder/++rest++url/nested1 HTTP/1.0')
    HTTP/1.0 200 OK
    Content-Length: 43
    Content-Type: text/html;charset=utf-8
    <BLANKLINE>
    http://localhost/folder/++rest++url/nested1

    >>> print http('GET /folder/++rest++url/nested2 HTTP/1.0')
    HTTP/1.0 200 OK
    Content-Length: 43
    Content-Type: text/html;charset=utf-8
    <BLANKLINE>
    http://localhost/folder/++rest++url/nested2


"""
from OFS.Folder import Folder
from five import grok
from infrae.rest import REST
from zope.traversing.browser import absoluteURL


class URLTest(REST):
    grok.context(Folder)
    grok.name('url')

    def GET(self):
        return absoluteURL(self, self.request)


class Nested1URLTest(REST):
    grok.adapts(URLTest, Folder)
    grok.name('nested1')

    def GET(self):
        return absoluteURL(self, self.request)


class Nested2URLTest(REST):
    grok.adapts(URLTest, Folder)
    grok.name('nested2')

    def GET(self):
        return absoluteURL(self, self.request)




"""

   Grok the test first:

   >>> grok('infrae.rest.tests.grok.nested')

  Now, if we have a folder we can use our REST component:

    >>> root = getRootFolder()
    >>> _ = root.manage_addProduct['OFS'].manage_addFolder('folder', 'Folder')
    >>> print http('GET /folder/++rest++testroot HTTP/1.0')
    HTTP/1.0 200 OK
    Content-Length: 4
    Content-Type: plain/text
    <BLANKLINE>
    Root

  We can access sub-handlers, but only going through the root handler:

    >>> print http('GET /folder/++rest++testroot/testchildfirst HTTP/1.0')
    HTTP/1.0 200 OK
    Content-Length: 11
    Content-Type: plain/text
    <BLANKLINE>
    Child First

    >>> print http('GET /folder/++rest++testchildfirst HTTP/1.0')
    Traceback (most recent call last):
       ...
    NotFound: testchildfirst

  And sub-handlers can have themselves sub-handlers:

    >>> print http('GET /folder/++rest++testroot/testchildfirst/testgrandchildfirst HTTP/1.0')
    HTTP/1.0 200 OK
    Content-Length: 17
    Content-Type: plain/text
    <BLANKLINE>
    Grand Child First

  Same way, sub-handlers can have security and a specific name:

    >>> print http('GET /folder/++rest++testroot/otherchild HTTP/1.0')
    HTTP/1.0 401 Unauthorized
    Content-Length: 0
    Www-Authenticate: basic realm="Zope"
    <BLANKLINE>
    <BLANKLINE>

"""

from OFS.Folder import Folder
from five import grok
from infrae.rest import REST


class TestRoot(REST):
    grok.context(Folder)

    def GET(self):
        self.response.setHeader('Content-Type', 'plain/text')
        return "Root"


class TestChildFirst(REST):
    grok.adapts(TestRoot, Folder)

    def GET(self):
        self.response.setHeader('Content-Type', 'plain/text')
        return "Child First"


class TestChildSecond(REST):
    grok.adapts(TestRoot, Folder)
    grok.name('otherchild')
    grok.require('zope2.ViewManagementScreens')

    def GET(self):
        self.response.setHeader('Content-Type', 'plain/text')
        return "Child Second"


class TestGrandChildFirst(REST):
    grok.adapts(TestChildFirst, Folder)

    def GET(self):
        self.response.setHeader('Content-Type', 'plain/text')
        return "Grand Child First"



# Copyright (c) 2010 Infrae. All rights reserved.
# See also LICENSE.txt
# $Id$

from five import grok
from zope.component import getGlobalSiteManager, getMultiAdapter
from zope.event import notify
from zope.interface import Interface, providedBy
from zope.interface.interfaces import IInterface
from zope.traversing.browser.interfaces import IAbsoluteURL
from zope.traversing.namespace import view

from Products.Five import BrowserView
from zExceptions import NotFound

from infrae.rest.interfaces import RESTMethodPublishedEvent
from infrae.rest.interfaces import MethodNotAllowed, IRESTComponent

import json

ALLOWED_REST_METHODS = ('GET', 'POST', 'HEAD', 'PUT',)


def queryRESTComponent(specs, args, name=u''):
    """Query the ZCA for a REST component.
    """
    def specOf(obj):
        if IInterface.providedBy(obj):
            return obj
        return providedBy(obj)

    sm = getGlobalSiteManager()
    factory = sm.adapters.lookup(map(specOf, specs), IRESTComponent, name)
    if factory is not None:
        result = factory(*args)
        if result is not None and IRESTComponent.providedBy(result):
            return result
    return None


def lookupREST(context, request, name):
    """Lookup a REST component called name on the given context / request.
    """
    view = queryRESTComponent(
        (Interface, context),
        (context, request),
        name=name)
    if view is None:
        raise NotFound(name)
    # Set view parent/name for security
    view.__name__ = name
    view.__parent__ = context
    return view


class REST(object):
    """A base REST component
    """
    grok.baseclass()
    grok.implements(IRESTComponent)

    def __init__(self, context, request):
        self.context = context
        self.request = request

    @property
    def response(self):
        return self.request.response

    def browserDefault(self, request):
        """Render the component using a method called the same way
        that the HTTP method name.
        """
        if request.method in ALLOWED_REST_METHODS:
            if hasattr(self, request.method):
                return self, (request.method,)
        raise MethodNotAllowed(request.method)

    def publishTraverse(self, request, name):
        """You can traverse to a method called the same way that the
        HTTP method name, or a sub view
        """
        if name in ALLOWED_REST_METHODS and name == request.method:
            if hasattr(self, name):
                notify(RESTMethodPublishedEvent(self, name))
                return getattr(self, name)
        view = queryRESTComponent(
            (self, self.context),
            (self.context, request),
            name=name)
        if view is not None:
            # Set parenting information
            view.__name__ = name
            view.__parent__ = self
            return view
        raise NotFound(name)

    def json_response(self, result):
        """Encode a result as a JSON response.
        """
        self.response.setHeader('Content-Type', 'application/json')
        return json.dumps(result)


class RESTWithTemplate(REST):
    template = None

    def default_namespace(self):
        return {'rest': self,
                'context': self.context,
                'request': self.request}

    def namespace(self):
        return {}


class MethodNotAllowedView(grok.View):
    grok.context(MethodNotAllowed)
    grok.name('error.html')

    def update(self):
        self.response.setStatus(405)

    def render(self):
        return u"Method not allowed"


class RESTNamespace(view):
    """Implement a namespace ++rest++.
    """

    def traverse(self, name, ignored):
        self.request.shiftNameToApplication()
        if name:
            return lookupREST(self.context, self.request, name)
        return self.context


class AbsoluteURL(BrowserView):

    def __init__(self, context, request):
        self.context = context
        self.request = request
        self._parent_url = getMultiAdapter(
            (context.__parent__, request), IAbsoluteURL)

    def __str__(self):
        pattern = '%s/++rest++%s'
        if IRESTComponent.providedBy(self.context.__parent__):
            pattern = '%s/%s'
        return pattern % (self._parent_url(), self.context.__name__)

    __call__ = __repr__ = __unicode__ = __str__

    def breadcrumbs(self):
        return self._parent_url.breadcrumbs()



# Copyright (c) 2010 Infrae. All rights reserved.
# See also LICENSE.txt
# $Id$


from zope.publisher.interfaces.browser import IBrowserPublisher
from zope.location.interfaces import ILocation
from zope.browser.interfaces import IBrowserView
from zope.component.interfaces import IObjectEvent, ObjectEvent
from zope.interface import Attribute, implements


class IRESTComponent(IBrowserPublisher, ILocation, IBrowserView):
    """A REST component
    """


class IRESTMethodPublishedEvent(IObjectEvent):
    """Event triggered when a method is published.
    """
    method = Attribute("Published method name")


class RESTMethodPublishedEvent(ObjectEvent):
    implements(IRESTMethodPublishedEvent)

    def __init__(self, object, method):
        super(RESTMethodPublishedEvent, self).__init__(object)
        self.method = method


class MethodNotAllowed(Exception):
    """This method is not allowed.
    """




"""

  Let's first grok our test first:

   >>> grok('infrae.rest.tests.grok.template')

  Now, if we have a folder we can use our REST component:

    >>> root = getRootFolder()
    >>> root.manage_addProduct['OFS'].manage_addFolder('folder', 'Folder')

  We can now fetch the component with the template:

    >>> print http('GET /folder/++rest++form HTTP/1.0')
    HTTP/1.0 200 OK
    Content-Length: 30
    Content-Type: text/html;charset=utf-8
    <BLANKLINE>
    <form>
       <h1>42</h1>
    </form>
    <BLANKLINE>

  And the nested one as well:

    >>> print http('GET /folder/++rest++form/widget HTTP/1.0')
    HTTP/1.0 200 OK
    Content-Length: 15
    Content-Type: text/html;charset=utf-8
    <BLANKLINE>
    <span>51</span>


"""

from OFS.Folder import Folder
from five import grok
from infrae.rest import RESTWithTemplate


class FormTemplate(RESTWithTemplate):
    grok.context(Folder)
    grok.name('form')


    def value(self):
        return '42'

    def GET(self):
        return self.template.render(self)


class WidgetTemplate(RESTWithTemplate):
    grok.adapts(FormTemplate, Folder)
    grok.name('widget')

    def value(self):
        return '51'

    def GET(self):
        return self.template.render(self)


# Let's use an inline template here
widgettemplate = grok.PageTemplate('<span tal:content="rest/value"></span>')



# Copyright (c) 2010 Infrae. All rights reserved.
# See also LICENSE.txt
# $Id$

from App.class_init import InitializeClass as initializeClass
from Products.Five.security import protectName

from five import grok
from grokcore.view.meta.views import TemplateGrokker
from zope.interface import Interface
from zope.component import adaptedBy, provideAdapter
import martian

from infrae.rest.components import REST, ALLOWED_REST_METHODS
from infrae.rest.components import RESTWithTemplate
from infrae.rest.interfaces import IRESTComponent

def default_view_name(factory, module=None, **data):
    return factory.__name__.lower()

def default_context(*stuff, **things):
    return None


class RESTGrokker(martian.ClassGrokker):
    """Grok REST views.
    """
    martian.component(REST)
    martian.directive(grok.context, get_default=default_context)
    martian.directive(grok.name, get_default=default_view_name)
    martian.directive(grok.require, name='permission')

    def execute(self, factory, context, name, permission, config, **kw):
        """Register the REST component as a view on the IREST layer.
        """
        if permission is None:
            permission = 'zope.Public'

        adapts = adaptedBy(factory)
        if adapts is None:
            adapts = (Interface, context)

        config.action(
            discriminator=('adapter', adapts, IRESTComponent, name),
            callable=provideAdapter,
            args=(factory, adapts, IRESTComponent, name))

        for method in ALLOWED_REST_METHODS:
            if hasattr(factory, method):
                config.action(
                    discriminator = ('five:protectName', factory, method),
                    callable = protectName,
                    args = (factory, method, permission))

        config.action(
            discriminator = ('five:initialize:class', factory),
            callable = initializeClass,
            args = (factory,))
        return True


class RESTWithTemplateGrokker(TemplateGrokker):
    martian.component(RESTWithTemplate)

    def has_render(self, factory):
        return False

    def has_no_render(self, factory):
        return False

